@@ -45,7 +45,8 @@
 #include "qqmlobjectcreator_p.h"
 #include <private/qqmlcomponent_p.h>
 
-void QQmlEnginePrivate::incubate(QQmlIncubator &i, QQmlContextData *forContext)
+void QQmlEnginePrivate::incubate(
+        QQmlIncubator &i, const QQmlRefPointer<QQmlContextData> &forContext)
 {
     QExplicitlySharedDataPointer<QQmlIncubatorPrivate> p(i.d);
 
@@ -59,13 +60,13 @@ void QQmlEnginePrivate::incubate(QQmlIncubator &i, QQmlContextData *forContext)
 
         // Need to find the first constructing context and see if it is asynchronous
         QExplicitlySharedDataPointer<QQmlIncubatorPrivate> parentIncubator;
-        QQmlContextData *cctxt = forContext;
+        QQmlRefPointer<QQmlContextData> cctxt = forContext;
         while (cctxt) {
-            if (!cctxt->hasExtraObject && cctxt->incubator) {
-                parentIncubator = cctxt->incubator;
+            if (QQmlIncubatorPrivate *incubator = cctxt->incubator()) {
+                parentIncubator = incubator;
                 break;
             }
-            cctxt = cctxt->parent;
+            cctxt = cctxt->parent();
         }
 
         if (parentIncubator && parentIncubator->isAsynchronous) {
@@ -149,8 +150,8 @@ void QQmlIncubatorPrivate::clear()
     }
     enginePriv = nullptr;
     if (!rootContext.isNull()) {
-        if (!rootContext->hasExtraObject)
-            rootContext->incubator = nullptr;
+        if (rootContext->incubator())
+            rootContext->setIncubator(nullptr);
         rootContext = nullptr;
     }
 
@@ -268,6 +269,7 @@ void QQmlIncubatorPrivate::forceCompletion(QQmlInstantiationInterrupt &i)
     }
 }
 
+
 void QQmlIncubatorPrivate::incubate(QQmlInstantiationInterrupt &i)
 {
     if (!compilationUnit)
@@ -279,6 +281,20 @@ void QQmlIncubatorPrivate::incubate(QQmlInstantiationInterrupt &i)
     // get a copy of the engine pointer as it might get reset;
     QQmlEnginePrivate *enginePriv = this->enginePriv;
 
+    // Incubating objects takes quite a bit more stack space than our usual V4 function
+    enum { EstimatedSizeInV4Frames = 2 };
+    QV4::ExecutionEngineCallDepthRecorder<EstimatedSizeInV4Frames> callDepthRecorder(
+                compilationUnit->engine);
+    if (callDepthRecorder.hasOverflow()) {
+        QQmlError error;
+        error.setMessageType(QtCriticalMsg);
+        error.setUrl(compilationUnit->url());
+        error.setDescription(QQmlComponent::tr("Maximum call stack size exceeded."));
+        errors << error;
+        progress = QQmlIncubatorPrivate::Completed;
+        goto finishIncubate;
+    }
+
     if (!vmeGuard.isOK()) {
         QQmlError error;
         error.setMessageType(QtInfoMsg);
@@ -360,10 +376,8 @@ void QQmlIncubatorPrivate::incubate(QQmlInstantiationInterrupt &i)
             if (watcher.hasRecursed())
                 return;
 
-            QQmlContextData *ctxt = nullptr;
-            ctxt = creator->finalize(i);
-            if (ctxt) {
-                rootContext = ctxt;
+            if (creator->finalize(i)) {
+                rootContext = creator->rootContext();
                 progress = QQmlIncubatorPrivate::Completed;
                 goto finishIncubate;
             }
@@ -410,26 +424,6 @@ void QQmlIncubationController::incubateFor(int msecs)
     } while (d && d->incubatorCount != 0 && !i.shouldInterrupt());
 }
 
-#if QT_DEPRECATED_SINCE(5, 15)
-/*!
-\obsolete
-
-\warning Do not use this function.
-Use the overload taking a \c{std::atomic<bool>} instead.
-*/
-void QQmlIncubationController::incubateWhile(volatile bool *flag, int msecs)
-{
-    if (!d || !d->incubatorCount)
-        return;
-
-    QQmlInstantiationInterrupt i(flag, msecs * Q_INT64_C(1000000));
-    i.reset();
-    do {
-        static_cast<QQmlIncubatorPrivate*>(d->incubatorList.first())->incubate(i);
-    } while (d && d->incubatorCount != 0 && !i.shouldInterrupt());
-}
-#endif
-
 /*!
 \since 5.15
 
